###########
# COLORS  #
###########

# Fill with theme variables later

###########
# STYLES  #
###########

#### MAIN CONTENT

CONTENT_STYLE = {
    "margin-left": "18rem",
    "margin-right": "2rem",
    "padding": "2rem 1rem",
}


#### SIDEBAR

SIDEBAR_STYLE = {
    "position": "fixed",
    "top": 0,
    "left": 0,
    "bottom": 0,
    "width": "18rem",
    "padding": "2rem 1rem",
    "overflow": "auto",
    "background-color": "#f8f9fa",
}


#### TAB STYLES

TAB_HEADER_STYLE = {
    "text-align": "center"
}

#### PLOT TABLES


PLOT_GROUP_STYLE = {
    "padding": "10px",
    "display": "grid",
    
    "border-width": "2px",
    "border-style": "solid",
    "border-color": "black",
    
    "overflow": "auto"
}

PLOT_GROUP_HEADER_STYLE = {
    "text-align": "center"
}


CONTENT_GRID_STYLE = {
    "padding": "10px",
    "display": "grid",

    "border-width": "2px",
    "border-style": "solid",
    "border-color": "black",

    "overflow": "auto"
}

CONTENT_GRID_NO_BORDER_STYLE = {
    "padding": "10px",
    "display": "grid",

    "overflow": "auto"
}

CONTENT_GRID_HEADER_STYLE = {
    "text-align": "center"
}

#### PANEL STYLES

PANEL_HEADER_STYLE = {
    "text-align": "center"
}

PANEL_STYLE = {
    "padding": "10px",

    "border-width": "2px",
    "border-style": "solid",
    "border-color": "black",

    "width": "100%"
}


#### PLOT STYLES

PLOT_FRAME_STYLE = {
    "padding": "10px",
    
    "border-width": "2px",
    "border-style": "solid",
    "border-color": "black",
}

PLOT_FRAME_HEADER_STYLE = {
    "text-align": "center"
}


CHECKLIST_STYLE = {
    "padding": "10px",
    
    "border-width": "1px",
    "border-style": "solid",
    "border-color": "black",
    
    "width": "200px",
    "height": "150px"
}



#### CONTROL STYLES

CONTROL_FRAME_STYLE = {
    "padding": "10px",
    
    "border-width": "1px",
    "border-style": "solid",
    "border-color": "black",
    
    # "height": "200px",
    # "position": "relative"
}


CONTROL_STYLE = {
    "padding": "5px",

    "border-width": "1px",
    "border-style": "solid",
    "border-color": "black"
}


PLOT_CONTROL_STYLE = {
    "padding": "5px",
    
    "border-width": "1px",
    "border-style": "solid",
    "border-color": "black",
}


# Delete after refactor
CONTROL_OBJECT_STYLE = {
    "padding": "5px",

    "border-width": "1px",
    "border-style": "solid",
    "border-color": "black"
}



import os

from dash import html
from dash import dcc

from quickboard.dashsetup import app
from quickboard.utils.environment import isnotebook


def start_app(board, mode='external', host=os.getenv("HOST", "127.0.0.1"), port=8050, extra_stores=[]):
    """
    Takes an app instance and configures its board layout, then runs the app on given port. Use extra_stores for
    extra data stores.
    """
    app.layout = html.Div([
        board.container,
        dcc.Store(id='sidebar_state', data={}),
        dcc.Store(id='data_store', data={}),
    ] + [
        dcc.Store(id=name, data={}) for name in extra_stores
    ])

    if isnotebook():
        app.run_server(mode=mode, debug=True, host=host, port=port)
    else:
        app.run_server(debug=True, host=host, port=port)



from quickboard.plugins import DataFilterChecklist


class DataSumChecklist(DataFilterChecklist):
    # Overrides parent method
    def data_transform(self, df, control_value):
        """
        Transforms data so that given column becomes the sum of the values given by `col-name_checklist-item` where
        checklist-item is currently checked.
        """
        df[self.data_col] = 0
        for cv in control_value:
            df[self.data_col] += df[self.data_col + '_' + cv]

        return df



from IPython import get_ipython

def isnotebook():
    """
    A function for determining if the program is being run in a notebook or command line.
    """
    try:
        shell = get_ipython().__class__
        if hasattr(shell, '__name__') & (shell.__name__ == 'ZMQInteractiveShell'):
            return True
        else:
            return False
    except NameError:
        return False



from dash import dcc
from dash import html
import dash_bootstrap_components as dbc

import pandas as pd

import quickboard.styles as styles
from quickboard.base import ContentGrid


class Panel:
    """
    An "abstract" class representing a panel holding a single Dash/HTML object.
    Inputs:
        header = header text/object
        main_content = main HTML object to hold
        kwargs = extra keyword arguments become attributes of the object for extending functionality easily
    """
    def __init__(self, header, main_content, **kwargs):
        self.header = html.H3(header, style=styles.PANEL_HEADER_STYLE) if type(header) == str else header
        self.main_content = main_content

        self.container = dbc.Toast([
            self.header, self.main_content
        ],
            style=styles.PANEL_STYLE
        )

        for key, value in kwargs.items():
            setattr(self, key, value)


class DynamicPanel(Panel):
    """
    A template for creating larger panels holding objects which can be updated via different controls and plugins.
    Inputs:
        header = header text/object
        main_content = main HTML object to hold
        data_source = key to use in tab data dictionary to get data inputs for this panel
        body = text/objects to present between header and main_content
        plugins = list of plugin objects to load under main_content to use to manipulate main object
        plugin_wrap = number of plugins to load per row underneath main object
        kwargs = extra keyword arguments become attributes of the object for extending functionality easily
    """
    def __init__(self, header, main_content, data_source="data", body="", plugins=[], plugin_wrap=2, **kwargs):
        super().__init__(header, main_content, **kwargs)
        self.data_source = data_source

        # Keep list of self as DynamicPanel to register it with larger objects holding it for tab-level callbacks
        self.dps = [self]

        # Add optional body text above main content under header
        self.body = html.Div([body, html.Br()])

        self.main_content = html.Div([main_content], style={'border-width': '1px', 'border-style': 'solid',
                                                            'border-color': 'black'})

        # Configure plugin related attributes
        self.plugins = plugins
        border = False if len(plugins) == 0 else True
        self.plugin_frame = ContentGrid(header="", entities=plugins, col_wrap=plugin_wrap, border=border)

        self.container = html.Div([
            self.header, self.body, html.Br(), self.main_content, html.Br(), self.plugin_frame.container
        ],
            style=styles.PANEL_STYLE
        )

    def data_transform(self, df):
        """
        A method for applying specific transformations to the data source before passing to main object, regardless
        of control object states.
        """
        return df

    def apply_transforms(self, data_state, control_values=[]):
        """
        A method that is called when relevant control objects change state to manipulate data source before
        passing to the main object.
        """
        df = pd.DataFrame.from_records(data_state[self.data_source])
        df = self.data_transform(df)

        # Find control plugin objects corresponding to the controls for this plot panel
        # control_html_ids = [x['html_id'] for x in control_ids]
        controls = [plugin for plugin in self.plugins if hasattr(plugin, 'control')]

        # Apply control plugin effects
        for control, control_value in zip(controls, control_values):
            df = control.configure(self, df, control_value)

        return df



from dash import dcc

from quickboard.base import ControlPlugin


class DataFilterRadioButtons(ControlPlugin):
    """
    A plugin for filtering data to be displayed by removing records where a certain column's value does not equal the
    selected radio button value.
    Inputs:
        header = header text/object
        data_col = column from data to check for values matching button
        data_values = list of possible values to populate the radio button list
    """
    def __init__(self, header, data_col, data_values):
        self.data_col = data_col
        component = dcc.RadioItems
        component_inputs = {
            'options': [
                {'label': x, 'value': x} for x in data_values
            ],
            'value': data_values[0],
            'labelStyle': {'display': 'block'},
            'inputStyle': {"margin-right": '10px'}
        }

        super().__init__(
            header=header,
            component=component,
            component_inputs=component_inputs
        )

    # Overrides parent method
    def data_transform(self, df, control_value):
        """
        Filters the data by checking the column has value matching the selected button.
        """
        df = df[df[self.data_col] == control_value]
        return df



from dash import dcc
from dash import html
import dash_bootstrap_components as dbc

from quickboard.dashsetup import app
import quickboard.styles as styles


class BaseTab:
    """
    The basic class for creating tabs in an app.
    Inputs:
        tab_label = name appearing on tab button, and used to reference tab elsewhere
        tab_header = opening text at top of tab page
        content_list = list of objects to display on the tab
        sidebar_header = header to use on top of sidebar while on this tab
        sidebar_plugins = plugins to use in the sidebar while on this tab
    """
    def __init__(self, tab_label, tab_header, content_list, sidebar_header="Data Controls", sidebar_plugins=[]):
        self.tab_label = tab_label
        self.tab = dcc.Tab(value=tab_label, label=tab_label)
        self.sidebar_header = sidebar_header
        self.sidebar_plugins = sidebar_plugins

        for plugin in self.sidebar_plugins:
            if hasattr(plugin, 'control'):
                plugin.control.id = {
                    'control_type': 'sidebar_control',
                    'unique_id': id(plugin)
                }

        # Collect dynamic panels from all subobjects into one place
        self.dps = []
        for entity in content_list:
            if hasattr(entity, 'dps'):
                self.dps += entity.dps

        content_containers = [x.container for x in content_list]
        for i in range(len(content_containers) - 1):
            content_containers.insert(2 * i + 1, html.Br())

        self.container = html.Div(
            [html.H1(tab_header, style=styles.TAB_HEADER_STYLE)] + content_containers,
        )



import dash
import jupyter_dash
import dash_bootstrap_components as dbc
from quickboard.utils.environment import isnotebook


theme = dbc.themes.BOOTSTRAP
if isnotebook():
    # Initialize app
    # Other nice themes: DARKLY, CYBORG, BOOTSTRAP, FLATLY, LUX, LUMEN, SOLAR
    app = jupyter_dash.JupyterDash(external_stylesheets=[theme])
    app.config.suppress_callback_exceptions = True   # Ignore errors from other tab features not present on start tab
else:
    app = dash.Dash(external_stylesheets=[theme])



from dash import html
from dash.dependencies import Input, Output, State, ALL

from quickboard.dashsetup import app

import pandas as pd


class GetUpdatedText:
    """
    A textbox that can be populated with the value of another dynamic object's state. Used to make dynamic text
    descriptions.
    Inputs:
        target_component = dynamic object to listen to
        target_prop = property of above object to listen for
        start_text = static text to use before the dynamic text
        end_text = static text to use after the dynamic text
    """
    def __init__(self, target_component, target_prop, start_text, end_text):
        self.textbox = html.Div([])
        self.start_text = start_text
        self.end_text = end_text

        app.callback(
            Output(self.textbox, 'children'),
            Input(target_component, target_prop)
        )(self.get_update)

    def get_update(self, property):
        return f"{self.start_text}{property}{self.end_text}"


class GetDataTableSize(GetUpdatedText):
    """
    A special case of GetUpdatedText to produce the size of a dynamic DataTable.
    """
    def __init__(self, target_component, start_text, end_text):
        target_prop = 'data'
        super().__init__(target_component, target_prop, start_text, end_text)

    def get_update(self, data):
        df = pd.DataFrame.from_records(data)
        return f"{self.start_text}{len(df)}{self.end_text}"


class GetUpdatedControlValue(GetUpdatedText):
    """
    A special case of GetUpdatedText to produce the value of given object (e.g. checklist, radio buttons, etc).
    """
    def __init__(self, target_component, start_text, end_text):
        target_prop = 'value'
        super().__init__(target_component, target_prop, start_text, end_text)



from dash import dcc
from dash import html
import dash_bootstrap_components as dbc

import quickboard.styles as styles

import itertools


class ContentGrid:
    """
    An object for storing other components into a grid.
    Inputs:
        header = header text/object
        entities = list of objects to store in the grid
        col_wrap = number of entities to display within a row
        border = boolean to determine if the grid has a surrounding border
    """
    def __init__(self, header, entities, col_wrap=2, border=True):
        self.header = html.H2(header, style=styles.CONTENT_GRID_HEADER_STYLE) if type(header) == str else header

        # Collect dynamic panels from all subobjects into one place
        self.dps = []
        for entity in entities:
            if hasattr(entity, 'dps'):
                self.dps += entity.dps

        table_rows = []
        table_entry_width = f'{round(100 / col_wrap, 2)}%'
        # Break up plot list into equal sized chunks with last row filled with "None" for spillovers
        for row in itertools.zip_longest(*(iter(entities),) * col_wrap):
            table_row = html.Tr([
                html.Td(
                    [x.container],
                    style={"width": "100%", "padding": "5px"}
                ) for x in row if x is not None
            ])
            table_rows.append(table_row)

        self.container = html.Div([
            self.header,
            html.Table(table_rows, style={"width": "100%", 'table-layout': 'fixed'})
        ],
            style=styles.CONTENT_GRID_STYLE if border else styles.CONTENT_GRID_NO_BORDER_STYLE
        )



from dash import dcc
from dash import html
import dash_bootstrap_components as dbc
from dash.dependencies import Input, Output, State, ALL

import pandas as pd

from quickboard.base.sidebar import Sidebar
from quickboard.dashsetup import app
import quickboard.styles as styles


class Quickboard:
    """
    Main class for making an easy dashboard out of modular components. Handles some global dynamic aspects of the
    board while holding all of the pieces together.
    Inputs:
        sidebar_header = header text/object to use if no tabs
        sidebar_plugins = list of plugins to use in sidebar if no tabs
        tab_list = list of tab objects from which the board is comprised
        global_contents = objects to display in the absence of tabs
        data_paths = dictionary of `{tab label: paths}` where `paths` is either string or dictionary of
                    `{data source name: path}` to be reference in dynamic panels within the given tab
    """
    def __init__(self, sidebar_header="Data Controls", sidebar_plugins=[], tab_list=[], global_contents=[],
                 data_paths={}):
        # Parse data paths
        self.data_paths = {}
        if isinstance(data_paths, str):
            for tab in tab_list:
                self.data_paths[tab.tab_label] = data_paths
        elif isinstance(data_paths, dict):
            self.data_paths = data_paths
        else:
            print("ERROR: Data paths must be either const string or dictionary of tab_label -> paths.")

        if len(tab_list) != 0:
            first_tab = tab_list[0]
            self.sidebar = Sidebar(first_tab.sidebar_header, first_tab.sidebar_plugins)
        else:
            self.sidebar = Sidebar(sidebar_header, sidebar_plugins)

        # Collect tabs together unless user inputs none
        self.tab_list = tab_list
        if len(tab_list) != 0:
            self.tabs = dcc.Tabs(
                    value=self.tab_list[0].tab_label,
                    children=[x.tab for x in self.tab_list]
            )

            self.current_tab_content = html.Div(children=html.P('If this message persists, then there was an ERROR '
                                                                'initializing tabs!'))

            self.tabs_container = html.Div(
                children=[
                    self.tabs,
                    self.current_tab_content
                ]
            )

            self.current_tab = self.tab_list[0]
            self.tab_dict = {
                tab.tab_label: tab for tab in tab_list
            }

        else:
            self.tabs_container = html.Div([])

        # Used in case user doesn't want to have tabs, but one page with global contents
        self.global_contents = html.Div(global_contents)

        self.container = html.Div(
            children=[
                self.sidebar.container,
                self.global_contents,
                self.tabs_container
            ],
            style=styles.CONTENT_STYLE
        )

        #############
        # CALLBACKS #
        #############

        # Add callback for tab switching
        if len(tab_list) > 0:
            app.callback(
                Output(self.current_tab_content, 'children'),
                Output(self.sidebar.container, 'children'),
                Input(self.tabs, 'value')
            )(self.tab_switch_update)

        # Add callback for updating data from sidebar events
        app.callback(
            Output('data_store', 'data'),
            Input({'control_type': 'sidebar_control', 'unique_id': ALL}, 'value'),
            Input(self.tabs, 'value'),
        )(self.update_data)

    def set_tab(self, tab_name):
        """
        Callback method for updating the current tab, based on user click.
        """
        self.current_tab = self.tab_dict[tab_name]
        return self.current_tab.container

    def update_sidebar_layout(self, tab_name):
        """
        Callback method for updating the sidebar layout corresponding to the current tab.
        """
        plugins = self.current_tab.sidebar_plugins
        plugin_containers = [x.container for x in plugins]

        # Put hlines between plugins
        hlines = [(plugin, html.Hr()) for plugin in plugin_containers]
        sidebar_layout = [y for sublist in hlines for y in sublist][:-1]

        return self.sidebar.header + sidebar_layout

    def tab_switch_update(self, tab_name):
        set_tab_container = self.set_tab(tab_name)
        updated_sidebar_layout = self.update_sidebar_layout(tab_name)

        return [set_tab_container, updated_sidebar_layout]

    def update_data(self, control_values=[], tab_name=""):
        """
        Callback method handling changes in current tab data sources. Can be triggered by either:
            change in current tab;
            interacting with sidebar plugins.
        """
        new_data_paths = self.data_paths[tab_name]

        # If just one string is passed for tab data, convert to simple dict for downstream compatability
        if isinstance(new_data_paths, str):
            new_data_paths = {'data': new_data_paths}

        def df_from_path(path):
            if path == '':
                return pd.DataFrame({})
            elif path.split('.')[-1] == 'tsv':
                return pd.read_csv(path, sep='\t')
            else:
                return pd.read_csv(path)

        # Read in dfs relevant to the current tab
        new_dfs = {x: df_from_path(new_data_paths[x]) for x in new_data_paths}

        # Find control plugin objects corresponding to the controls for the sidebar
        controls = [plugin for plugin in self.current_tab.sidebar_plugins if hasattr(plugin, 'control')]

        # Apply control plugin effects
        for control, value in zip(controls, control_values):
            for data_source in new_dfs:
                for dp in self.current_tab.dps:
                    new_dfs[data_source] = control.configure(dp, new_dfs[data_source], value)

        df_dicts = {x: new_dfs[x].to_dict('records') for x in new_dfs}
        return df_dicts



import dash
from dash import dcc
from dash import html
import dash_bootstrap_components as dbc

import quickboard.styles as styles


class Sidebar:
    """
    An HTML object for creating a sidebar on the left of the screen, used to control data and dynamic panels on the
    current tab page.
    Inputs:
        header_text = header text/object
        sidebar_plugins = plugins to use in the sidebar if there are no tabs in app
    """

    def __init__(self, header_text='Data Controls', sidebar_plugins=[]):
        self.header = [html.H2(header_text, className="display-4"), html.Hr()]
        self.plugins = sidebar_plugins

        self.container = html.Div(
            children=self.header + [plugin.container for plugin in self.plugins],
            style=styles.SIDEBAR_STYLE
        )



from .datadisplay import DataDisplay
from .datafilterchecklist import DataFilterChecklist
from .datafilterradiobuttons import DataFilterRadioButtons
from .datafilterslider import DataFilterSlider
from .datasumchecklist import DataSumChecklist
from .plotinputradiobuttons import PlotInputRadioButtons


from dash import dcc
from dash import dash_table
from dash.dependencies import Input, Output, State, ALL

from quickboard.dashsetup import app
from quickboard.base import DynamicPanel

import pandas as pd


class DataPanel(DynamicPanel):
    """
    A dynamic panel meant to hold a Dash DataTable.
    Inputs:
        header = header text/object
        data_source = key to use in tab data dictionary to get data inputs for this panel
        body = text/objects to present between header and main_content
        plugins = list of plugin objects to load under main_content to use to manipulate main object
        plugin_wrap = number of plugins to load per row underneath main object
        kwargs = extra keyword arguments become attributes of the object for extending functionality easily
    """
    def __init__(self, header, data_source="data", body="", plugins=[], plugin_wrap=2, **kwargs):
        self.datatable = dash_table.DataTable(
            page_action='none',
            sort_action='native',
            filter_action='native',
            fixed_rows={'headers': True},
            style_table={'overflow': 'auto', 'width': '100%', 'height': '275px'}
        )

        super().__init__(
            header=header,
            main_content=self.datatable,
            data_source=data_source,
            plugins=plugins,
            body=body,
            plugin_wrap=plugin_wrap,
            **kwargs
        )

        # Table update callback
        app.callback(
            Output(self.datatable, 'data'),
            Output(self.datatable, 'columns'),
            Input('data_store', 'data'),
            [Input(x.control, 'value') for x in self.plugins if hasattr(x, 'control')]
        )(self.update_table)

    def update_table(self, data_state, control_values=[]):
        """
        A method called to populate the table when the state of a control object is changed.
        """
        df = self.apply_transforms(data_state, control_values)

        data = df.to_dict('records')
        columns = [{'id': c, 'name': c} for c in df.columns]
        return data, columns



from dash import dcc
from dash import html
import dash_bootstrap_components as dbc
from dash.dependencies import Input, Output, State, ALL

import quickboard.styles as styles
from quickboard.dashsetup import app
from quickboard.base import DynamicPanel

import pandas as pd


# Class representing a plot object with surrounding HTML objects
class PlotPanel(DynamicPanel):
    """
    A dynamic panel meant to hold a Plotly figure.
    Inputs:
        header = header text/object
        plotter = function which outputs a Plotly figure
        plot_inputs = dictionary of inputs which is passed to plotter to produce figure
        data_source = key to use in tab data dictionary to get data inputs for this panel
        body = text/objects to present between header and main_content
        plugins = list of plugin objects to load under main_content to use to manipulate main object
        plugin_wrap = number of plugins to load per row underneath main object
        kwargs = extra keyword arguments become attributes of the object for extending functionality easily
    """
    def __init__(self, header, plotter, plot_inputs, data_source="data", body="", plugins=[], plugin_wrap=2, **kwargs):
        # Plot specific attributes
        self.plot_inputs = plot_inputs
        self.plotter = plotter
        self.graph = dcc.Graph()

        super().__init__(
            header=header,
            main_content=self.graph,
            data_source=data_source,
            plugins=plugins,
            body=body,
            plugin_wrap=plugin_wrap,
            **kwargs
        )

        # Plot update callback
        app.callback(
            Output(self.graph, 'figure'),
            Input('data_store', 'data'),
            [Input(x.control, 'value') for x in self.plugins if hasattr(x, 'control')]
        )(self.make_plot)

    def make_plot(self, data_state, *control_values):
        """
        A method called to create the figure when the state of a control object is changed.
        """
        df = self.apply_transforms(data_state, control_values)
        fig = self.plotter(df, **self.plot_inputs)
        return fig



from dash import dcc

from quickboard.base import ControlPlugin


class PlotInputRadioButtons(ControlPlugin):
    """
    A plugin for modifying a PlotPanel's plotting behavior. The `plot_input` value can be dynamically changed based on
    the selection via radio buttons.
    Inputs:
        header = header text/object
        plot_input = name of the PlotPanel's plotter input to be changed by clicking the radio buttons
        data_values = list of possible values to populate the radio button list
    """
    def __init__(self, header, plot_input, data_values):
        self.plot_input = plot_input
        component = dcc.RadioItems
        component_inputs = {
            'options': [
                {'label': x, 'value': x} for x in data_values
            ],
            'value': data_values[0],
            'labelStyle': {'display': 'block'},
            'inputStyle': {"margin-right": '10px'}
        }

        super().__init__(
            header=header,
            component=component,
            component_inputs=component_inputs
        )

    # Overrides parent method
    def panel_transform(self, dp, control_value):
        """
        Changes the PlotPanel's plotting behavior to reflect the user choice.
        """
        dp.plot_inputs[self.plot_input] = control_value



from dash import html
from dash import dash_table
from dash.dependencies import Input, Output, State, ALL

from quickboard.dashsetup import app
from quickboard.base import Panel


class DataDisplay(Panel):
    """
    A plugin for showing data beneath a DynamicPanel, with listening capabilities.
    Inputs:
        header = header text/object
        data_source = key to use in tab data dictionary to get data inputs for this panel
        listen = list of control objects to get notified of changes in them
    """
    def __init__(self, header, data_source, listen=[], **kwargs):
        # Calibrate header based on input and control type
        if isinstance(header, str):
            header = html.H5(header)
        else:
            header = header

        self.data_source = data_source

        self.datatable = dash_table.DataTable(
            page_action='none',
            sort_action='native',
            filter_action='native',
            fixed_rows={'headers': True},
            style_table={'height': '150px', 'overflow': 'auto', 'overflowX': 'scroll', 'width': '100%'}
        )

        super().__init__(header=header, main_content=self.datatable, **kwargs)

        app.callback(
            Output(self.datatable, 'data'),
            Output(self.datatable, 'columns'),
            Input('data_store', 'data'),
            [Input(x, 'value') for x in listen]
        )(self.update_table)

    def data_transform(self, df):
        """
        A method for transforming the data before getting put into the table.
        """
        return df

    def update_table(self, data_state, *inputs):
        """
        Callback method for manipulating the data before getting put into the table. Use the `inputs` list to get
        the new states of the control objects declared in the __init__ `listen` list. Must return a tuple
        `(data, columns)` where `data` is a dict of records (e.g. df.to_dict('records')) and columns is a list of
        dictionaries with keys 'id' and 'name' (e.g. [{'id': c, 'name': c} for c in df.columns]).
        """
        df = pd.DataFrame.from_dict(data_state[self.data_source])
        df = self.data_transform(df)

        data = df.to_dict('records')
        columns = [{'id': c, 'name': c} for c in df.columns]
        return (data, columns)


from dash import dcc
from dash import html
import dash_bootstrap_components as dbc

import quickboard.styles as styles
from quickboard.base import Panel


class ControlPlugin(Panel):
    """
    Creates an HTML component representing controls for a DynamicPanel, as a plugin.
    Inputs:
        header = header text/object
        component = component from dcc or similar
        component_inputs = the inputs to set up the component
    """

    def __init__(self, header, component, component_inputs, **kwargs):

        # Calibrate header based on input
        if isinstance(header, str):
            header = html.H3(header)
        else:
            header = header

        self.control = component(**component_inputs)
        super().__init__(header=header, main_content=self.control, **kwargs)

    def data_transform(self, df, control_value):
        """
        A method for transforming the data to be used in the parent object; called when the control object is changed.
        """
        return df

    def panel_transform(self, dp, control_value):
        """
        A method for transforming the parent DynamicPanel to modify its attributes; called when the control object
        is changed.
        """
        pass

    def configure(self, dp, df, control_value):
        """
        Manipulates the DataFrame and DynamicPanel associated with the plugin according to the transform attributes of
        the plugin. Always returns transformed dataframe for further filtering without needing to serialize when passing
        between components, but modifies DPs "in place".
        """
        df = self.data_transform(df, control_value)
        self.panel_transform(dp, control_value)

        return df



from dash import dcc

from quickboard.base import ControlPlugin


class DataFilterSlider(ControlPlugin):
    """
    A plugin for filtering data to be displayed by removing records where a certain column's value does not lie in
    the range specified by the slider.
    Inputs:
        header = header text/object
        data_col = column from data to restrict values from slider; must be numerical
        slider_min = the minimal value of the slider
        slider_max = the maximal value of the slider
        slider_default_values = the value(s) the slider starts on by default; if a single number, then slider will be
            one-sided, and data will be restricted to be less than slider value; if a pair (tuple or list) of numbers,
            then slider will be two-sided, and data will be restricted to lie between the two values
        slider_step = controls how granular the slider will be; the minimal change in the slider value when dragging
        slider_marks = dictionary of the form `{num: label}` where position `num` on slider gets label `label`
        edges_infinite = if True, treat the edges of slider as infinite (no restriction on data when min/max)
    """
    def __init__(self, header, data_col, slider_min, slider_max, slider_default_values, slider_step,
                 slider_marks={}, edges_infinite=True, **kwargs):
        self.data_col = data_col
        self.slider_min = slider_min
        self.slider_max = slider_max
        self.edges_infinite = edges_infinite

        try:
            float(slider_default_values)
            component = dcc.Slider
        except TypeError:
            try:
                tuple(slider_default_values)
                if len(slider_default_values) == 2:
                    component = dcc.RangeSlider
                else:
                    print("ERROR: If using range of defaults, must input a pair.")
            except TypeError:
                print("ERROR: Default values must be either single number or pair of values in list/tuple.")

        # If no marks provided, use the min/max values as only labels
        if slider_marks == {}:
            slider_marks = {slider_min: str(slider_min), slider_max: str(slider_max)}

        component_inputs = {
            'min': slider_min,
            'max': slider_max,
            'value': slider_default_values,
            'step': slider_step,
            'included': True,
            'marks': slider_marks
        }

        super().__init__(
            header=header,
            component=component,
            component_inputs=component_inputs
        )

    def data_transform(self, df, control_value):
        """
        Filters data so given column lies within the range of the slider.
        """
        try:
            # Handle case of normal one-sided slider
            float(control_value)
            if self.edges_infinite & (control_value == self.slider_max):
                df = df
            else:
                df = df[df[self.data_col] <= control_value]
        except TypeError:
                try:
                    # Handle case of two-sided slider
                    tuple(control_value)
                    # Left side of slider
                    if self.edges_infinite & (control_value[0] == self.slider_min):
                        df = df
                    else:
                        df = df[df[self.data_col] >= control_value[0]]

                    # Right side of slider
                    if self.edges_infinite & (control_value[1] == self.slider_max):
                        df = df
                    else:
                        df = df[df[self.data_col] <= control_value[1]]
                except TypeError:
                    print("ERROR: Invalid control_value. Expecting single input or list/tuple of slider values.")

        return df
